class Categories:
    Job = "job"
    Monitor = "monitor"
    Notify = "notify"
    Webhook = "webhook"


class QueueNames:
    Job = "jobs"
    Monitor = "monitor"
    Notify = "notify"
    Webhook = "webhooks"
    Error = "error"
